#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "strhash.h"

#define PROBE_MAX            (6)
#define PROBE_FUNC(k,n,size) (((k)+(n))%(size))

void hash_dump(strtable *st)
{
  int key;

  HASH_BEGIN(st,key);

  tsm_debug("Dumping hash table %p, size %d\n",st,st->size);

  while(key!=HASH_END(st))
  {
    tsm_debug("%05d]\t\"%s\"(%u) -> %p\n",key,st->table[key].key,
      str_hash(st->table[key].key)%(st->size),st->table[key].output);

    HASH_NEXT(st,key);
  }
}

strtable *hash_create(int size)
{
  strtable *st;

  if(size<=0) return(NULL);

  st=(strtable *)malloc(sizeof(strtable));
  if(st==NULL) return(NULL);

  st->size=size;
  st->table=(strhash_entry *)malloc(sizeof(strhash_entry)*size);
  if(st->table==NULL)
  {
    free(st);
    return(NULL);
  }

  size--;
  while(size>=0)
  {
    st->table[size].key=NULL;
    st->table[size].output=NULL;
    size--;
  }

  return(st);
}

void hash_free(strtable *st)
{
  int n;

  if(st==NULL) return;

#ifdef DEBUG
  hash_dump(st);
#endif

  for(n=0; n<st->size; n++)
    if(st->table[n].key!=NULL)
      free((char *)st->table[n].key);

  free(st->table);
  free(st);
}

int hash_add_str(strtable *st, const char *str, void *value)
{
  unsigned int k=str_hash(str);
  int n;

  for(n=0; n<PROBE_MAX; n++)
  {
    unsigned int key=PROBE_FUNC(k,n,st->size);
    if(st->table[key].key==NULL)
    {
      st->table[key].key=strdup(str);
      st->table[key].output=value;
      return(0);
    }
  }

  return(-1);    
}

void *hash_remove_str(strtable *st, const char *str)
{
  unsigned int k=str_hash(str);
  int n;
  for(n=0; n<PROBE_MAX; n++)
  {
    unsigned int key=PROBE_FUNC(k,n,st->size);
    if(st->table[key].key == NULL) continue;
    if(strcmp(str,st->table[key].key)==0)
    {
      void *tmp=st->table[key].output;
      free((char *)st->table[key].key);
      st->table[key].key=NULL;
      return(tmp);
    }
  }

  return(NULL);

/*  unsigned int key=str_hash(str)%(st->size);
  if(st->table[key].key!=NULL)
    if(strcmp(str,st->table[key].key)==0)
    {
      void *tmp=st->table[key].output;
      free(st->table[key].key);
      st->table[key].key=NULL;
      return(tmp);
    }*/
//  return(NULL);
}

const char *hash_get_str(strtable *st, unsigned int k)
{
  int n;
  for(n=0; n<PROBE_MAX; n++)
  {
    unsigned int key=PROBE_FUNC(k,n,st->size);
    if(st->table[key].key == NULL) continue;
    if(str_hash(st->table[key].key) == k)
      return(st->table[key].key);
  }

  return(NULL);
//  key%=(st->size);
//  if(st->table[key].key ==NULL) return(NULL);
//  return(st->table[key].key);
}

void *hash_lookup_str(strtable *st, const char *str)
{
  unsigned int k=str_hash(str);
  int n;

  for(n=0; n<PROBE_MAX; n++)
  {
    unsigned int key=PROBE_FUNC(k,n,st->size);
    if(st->table[key].key!=NULL)
      if(str_hash(st->table[key].key) == k)
        return(st->table[key].output);
  }

  return(NULL);
}

void *hash_lookup_key(strtable *st, unsigned int k)
{
  int n;
  for(n=0; n<PROBE_MAX; n++)
  {
    unsigned int key=PROBE_FUNC(n,k,st->size);
    if(st->table[key].key == NULL) continue;
    if(k == str_hash(st->table[key].key)) return(st->table[key].output);
  }

  return(NULL);
//  unsigned int key=k%st->size;  
//  if(st->table[key].key!=NULL)
//    return(st->table[key].output);
//  return(NULL);
}

void *hash_remove_key(strtable *st, unsigned int k)
{
  int n;
  for(n=0; n<PROBE_MAX; n++)
  {
    unsigned int key=PROBE_FUNC(n,k,st->size);
    if(st->table[key].key == NULL) continue;
    if(k == str_hash(st->table[key].key))
    {
      void *tmp=st->table[key].output;
      free((char *)st->table[key].key);
      st->table[key].key=NULL;
      st->table[key].output=NULL;
      return(tmp);    
    }
  }

  return(NULL);
/*  k%=(st->size);
  if(st->table[k].key!=NULL)
  {
    void *tmp=st->table[k].output;
    free(st->table[k].key);
    st->table[k].key=NULL;
    st->table[k].output=NULL;
    return(tmp);    
  }*/
}

int hash_add_str_static(strtable *st, const char *str, void *value)
{
  unsigned int k=str_hash(str);
  int n;

  for(n=0; n<PROBE_MAX; n++)
  {
    unsigned int key=PROBE_FUNC(k,n,st->size);
    if(st->table[key].key==NULL)
    {
      st->table[key].key=str;
      st->table[key].output=value;
      return(0);
    }
  }

  return(-1);    
}

void *hash_remove_str_static(strtable *st, const char *str)
{
  unsigned int k=str_hash(str);
  int n;
  for(n=0; n<PROBE_MAX; n++)
  {
    unsigned int key=PROBE_FUNC(k,n,st->size);
    if(st->table[key].key == NULL) continue;
    if(strcmp(str,st->table[key].key)==0)
    {
      void *tmp=st->table[key].output;
      st->table[key].key=NULL;
      return(tmp);
    }
  }

  return(NULL);
}

void *hash_remove_key_static(strtable *st, unsigned int k)
{
  int n;
  for(n=0; n<PROBE_MAX; n++)
  {
    unsigned int key=PROBE_FUNC(n,k,st->size);
    if(st->table[key].key == NULL) continue;
    if(k == str_hash(st->table[key].key))
    {
      void *tmp=st->table[key].output;
      st->table[key].key=NULL;
      st->table[key].output=NULL;
      return(tmp);    
    }
  }

  return(NULL);
}


/*
int hash_add_key(strtable *st, unsigned int k, void *value)
{
  char buf[32];
  k%=(st->size);
  if(st->table[k].key!=NULL) return(-1);
  sprintf(buf,"%u",k);
  st->table[k].key=strdup(buf);
  st->table[k].output=value;
  return(0);
}*/

static __inline__ unsigned char reverse_bit_order(unsigned char in)
{
  unsigned char out=0,n;
  for(n=0; n<(sizeof(char)*8); n++)
    out|=((in>>n)&1)<<(((sizeof(char)*8)-1)-n);

  return(out);
}

unsigned int str_hash(const char *str)
{
  unsigned int out=0,n;

  for(n=0; str[n]!='\0'; n++)
  {
    unsigned int c=str[n];
    int shift=(n*23);
    shift+=reverse_bit_order(str[n]);
    shift%=24;
    out+=str[n];
    out^=(c<<shift);
  }

  return(out);
}

#ifdef STRHASH_TEST
int main(int argc, char *argv[])
{
  char buf[512];
  while(!feof(stdin))
  {
    if(fgets(buf,511,stdin)==NULL) continue;

    tsm_debug("0x%08x: %s",str_hash(buf),buf);
  }
  return(0);
}
#endif/*STRHASH_TEST*/
