#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "tsmforth.h"

#define ALLOC(type,num) (type *)calloc((size_t)(num),sizeof(type))

#ifdef TSMSTACK_TEST
int main(int argc, char *argv[])
{
  const char *d="HI GUYS ALJ AF MY FAXCE IS A ROTTORN BANNANA";
  tsm_stack *s=tsm_stack_create(1024);

  fprintf(stderr,"tsm_var=%d, tsm_stackdata=%d\n",sizeof(tsm_var),
    sizeof(tsm_stackdata));

  if(s==NULL)
  {
    fprintf(stderr,"Couldn't create stack\n");
    return(1);
  }

  fprintf(stderr,"Created stack\n");

  if(tsm_stack_push_string(s,d)<0)
  {
    fprintf(stderr,"Couldn't push\n");
    tsm_stack_free(s);
    return(2);
  }

  fprintf(stderr,"First push done\n");

  if(tsm_stack_push_string(s,d)<0)
  {
    fprintf(stderr,"Couldn't push\n");
    tsm_stack_free(s);
    return(2);
  }

  fprintf(stderr,"Second push done\n");

  {
    tsm_data_e type;
    void *d;

    if(tsm_stack_peek(s,&type,&d)<0)
    {
      fprintf(stderr,"Couldn't peek stack\n");
    }
    else
    {
      fprintf(stderr,"Stack peeked(%d,%p)\n",type,d);
      fprintf(stderr,"	%s\n",(const char *)d);
    }
  }

  if(tsm_stack_pop(s)<0)
  {
    fprintf(stderr,"Couldn't pop\n");
    tsm_stack_free(s);
    return(3);
  }

  fprintf(stderr,"Done first pop\n");

  {
    tsm_data_e type;
    void *d;

    if(tsm_stack_peek(s,&type,&d)<0)
    {
      fprintf(stderr,"Couldn't peek stack\n");
    }
    else
    {
      fprintf(stderr,"Stack peeked(%d,%p)\n",type,d);
      fprintf(stderr,"	%s\n",(const char *)d);
    }
  }

  if(tsm_stack_pop(s)<0)
  {
    fprintf(stderr,"Couldn't pop\n");
    tsm_stack_free(s);
    return(3);
  }

  fprintf(stderr,"Done second pop\n");

  fprintf(stderr,"Top is %p\n",s->top);

  fprintf(stderr,"All done, freeing\n");
  tsm_stack_free(s);
  fprintf(stderr,"Freed\n");
  return(0);
}
#endif/*TSMSTACK_TEST*/

tsm_stack *tsm_stack_create(int size)
{
  tsm_stack *s;

  if(size<=0) return(NULL);

  s=ALLOC(tsm_stack,1);
  if(s==NULL) return(NULL);

  s->base=ALLOC(unsigned char,size);
  s->top=(tsm_stackdata *)(s->base+size);
  s->size=size;
  s->left=size;
  s->depth=0;

  return(s);
}

void tsm_stack_free(tsm_stack *s)
{
  if(s!=NULL)
  {
    if(s->base!=NULL)
      free(s->base);
    free(s);
  }
}

int tsm_stack_push(tsm_stack *s, tsm_data_e type, const void *data)
{
  int dsize=tsm_var_size(type,data),lsize;
  if(dsize<=0) return(-1);
  lsize=align_up(dsize+sizeof(tsm_stackdata));

  if(dsize>(s->left)) return(-1);

  {
    char *ts=(char *)s->top;
    ts-=lsize;
    s->top=(tsm_stackdata *)ts;
  }

  s->top->type=type;
  s->top->size=dsize;
  memcpy(s->top->bdata,data,dsize);
  s->left-=lsize;
  s->depth++;
  return(0);
}

int tsm_stack_peek(tsm_stack *s, tsm_data_e *type, void **data)
{
  if(s->left == s->size) return(-1); // stack empty

  *type = s->top->type;
  *data = s->top->bdata;
  return(0);
}

int tsm_stack_pop(tsm_stack *s)
{
  int size;
  if(s->left == s->size) return(-1); // stack empty

  size=s->top->size;
  size=align_up(size+sizeof(tsm_stackdata));

  {
    char *ts=(char *)s->top;
    ts+=size;
    s->top=(tsm_stackdata *)ts;
  }

  s->left+=size;
  s->depth--;
  return(0);
}

int tsm_stack_getpop(tsm_stack *s, tsm_data_e *type, void **data)
{
  if(tsm_stack_peek(s,type,data)<0) return(-1);
  return(tsm_stack_pop(s));
}

int tsm_stack_getpop_require(tsm_stack *s, tsm_data_e type, void **data)
{
  tsm_data_e gtype;

  if(tsm_stack_getpop(s,&gtype,data)<0)
  {
    fprintf(stderr,"	Underflow\n");
    return(-1);
  }
  if(gtype==type) return(0);
  fprintf(stderr,"	%d != %d\n",gtype,type);
  return(-1);
}


void tsm_stack_print(tsm_stack *_s)
{
  void *data;
  tsm_data_e type;
  tsm_stack s=(*_s);  

  fprintf(stderr,"TOP(");

  while(tsm_stack_getpop(&s,&type,&data)>=0)
  {
    if((type <0)||(type >= type_max))
    {
      fprintf(stderr,"[INVALID TYPE!  CORRUPTON]\n");
      return;
    }

    switch(type)
    {
    case TFD_INT:
      fprintf(stderr,"%d ",*(int *)(data));
      break;
    case TFD_STRING:
      fprintf(stderr,"\"%s\" ",(char *)(data));
      break;
    default:
      fprintf(stderr,"[BAD/UNKNOWN] ");
      break;
    }
  }

  fprintf(stderr,"]BOTTOM\n");
}

int tsm_stack_dump(tsm_stack *s, const char *cmdstr,...)
{
  va_list ap;
  char str[512];
  int retval;

  va_start(ap,cmdstr);
    retval=vsprintf(str,cmdstr,ap);  
  va_end(ap);

  if(retval<=0)
    return(-1);
  else
  {
    long len=(s->size-s->left);
    FILE *fp=fopen(str,"wb");

    if(fp==NULL) return(-1);
    fwrite(s->top,len,1,fp);
    fclose(fp);
    return(0);
  }  
}
