/**
 * tsmstack.h -- defines structures and interface for the TSMforth stack
 */
#ifndef __TSMSTACK_H__
#define __TSMSTACK_H__

#include "tsmtypes.h"

/* Stack data header */
typedef struct tsm_stackdata
{
  tsm_data_e     type;
  int           size;
  unsigned char bdata[0];
} tsm_stackdata;

/* Dynamically-allocated stack structure */
typedef struct tsm_stack
{
  unsigned char *base;
  tsm_stackdata  *top;

  int size,left,depth;
} tsm_stack;


#ifdef __cplusplus
{
#endif

/* Dynamically allocates a stack of given size */
tsm_stack *tsm_stack_create(int size);
/* Deallocates a stack */
void      tsm_stack_free  (tsm_stack *s);
/* Pushes a value onto the top of the stack */
int       tsm_stack_push  (tsm_stack *s, tsm_data_e type, const void *data);
/* Peeks the value on the top of the stack, but does not remove it */
int       tsm_stack_peek  (tsm_stack *s, tsm_data_e *type, void **data);
/* Remove the top value of the stack */
int       tsm_stack_pop   (tsm_stack *s);
/* Removes the top of the stack and retrieves it's value */
int       tsm_stack_getpop(tsm_stack *s, tsm_data_e *type, void **data);
/* Removes the top of the stack, pops, returns error if wrong type */
int       tsm_stack_getpop_require(tsm_stack *s, tsm_data_e type, void **data);
/* Dumps raw stack values to given filename, i.e.
 * tsm_stack_dump("dump%d.bin",3);
 */
int       tsm_stack_dump  (tsm_stack *s, const char *cmdstr,...);
/* Prints debugging output on stack */
void      tsm_stack_print(tsm_stack *s);


/* Returns nonzero if given stack is empty */
#define tsm_stack_empty(s) ((s)->top == NULL)

#ifdef __cplusplus
}
#endif
#endif/*__TSMSTACK_H__*/
