#ifndef __TSMTYPES_H__
#define __TSMTYPES_H__

#define IF_STR    "IF"
#define THEN_STR  "THEN"
#define ELSE_STR  "ELSE"
#define DO_STR    "DO"
#define WHILE_STR "WHILE"

typedef enum tsm_data_e
{
  /* Beginning of standard builtins */
  TFD_INT,
  TFD_STRING,
  TFD_INSTRUCTION,
  TFD_BUILTIN_MAX,

  /* Note that there may be more not listed here */
  TFD_EOF=-1
} tsm_data_e;

typedef enum tsm_inst_e
{
  TSM_INVALID,
  TSM_IF,
  TSM_IFN,
  TSM_JUMP,
  TSM_CALL,
  TSM_RET,
  TSM_SET_LOCAL,
  TSM_GET_LOCAL,
  TSM_BUILTIN_MAX
  /* Others may come after here, but aren't recorded in enum */
} tsm_inst_e;

typedef struct tsm_codedata
{
  tsm_data_e type;
  union
  {
    int       intvalue;
    tsm_inst_e instrvalue;
    char      strvalue[0];
  } value;
} tsm_codedata;

typedef struct var_handler
{
  int fixed;
  long (*fn)(const void *);
  void (*print)(const void *);
  int (*tostr)(char *buf,const void *);
} var_handler;

#define tsm_get_var_key(buf,key) sprintf(buf,"%x",key);

#ifdef __cplusplus
extern "C" {
#endif

extern const var_handler vhandlers[];
extern const int type_max;

long tsm_var_size(const tsm_data_e type, const void *ptr);

#ifdef __cplusplus
}
#endif

#endif/*__TSMTYPES_H__*/
